/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.clientlib.parse;

import cricketdaemon.Util;
import java.lang.reflect.*;
import java.util.*;

/**
 * Contains facility routines for implementing a new "ParseHandlerMap".
 * A XXXParseHandlerMap contains routines to parse the value field of type
 * XXX.  
 *
 * The ParseHandlerMap uses a design pattern that makes it easy for
 * a developer to parse a new field type XXX.  To acheive this, the 
 * ParseHandlerMap uses Reflections, which basically allow a 
 * XXXParseHandlerMap to register "function pointers" into a hashtable
 * provided by ParseHandlerMap.  The "function pointers" point to
 * a check handler and a parse handler method, and they are register 
 * with the ASCII key that corresponds to the ASCII name of the field type.
 *
 * Thus, to parse a new XXX type, one extends ParseHandlerMap and registers
 * the check handler method "check_XXX" and a parse handler method 
 * "parse_xxx" using the key "XXX".  Whenever the "XXX" field is encountered, 
 * the "check_xxx" method will be invoked to verify that a correct 
 * <code>dataStructure</code> has been passed into 
 * <code>Field.parseField()</code>. Next, the "parse_xxx" will be "invoked" 
 * to parse the value of the field.
 *
 * @author Allen Miu
 */
abstract class ParseHandlerMap 
{
    private Hashtable parseMap;
    private Hashtable checkMap;
    protected Class TYPE;

    /**
     * Constructs a new ParseHandlerMap.
     */
    protected ParseHandlerMap()
    {
	parseMap = new Hashtable();
	checkMap = new Hashtable();
	TYPE = getClass();
    }

    private Class[] parseHandlerArgType = 
    { (new Object()).getClass(), (new String()).getClass() };

    private Class[] checkHandlerArgType = 
    { (new Object()).getClass() };

    private Method getParseHandler(String type)
    {
	return (Method) parseMap.get(type);
    }

    private Method getTargetDataTypeChecker(String type)
    {
	return (Method) checkMap.get(type);
    }

    /**
     * Registers a "check_type" method and a "parse_type"
     * method for the specified <code>type</code>.  The subclass
     * is expected to implemented both methods 
     *
     * <code>public boolean check_type(Object dataStructure)</code> and
     * <code>public void parse_type(Object dataStructure, String value)</code>.
     *
     * The former method is used to ensure that the correct target 
     * data structure is passed into the <code>Field.parseField()</code>
     * while the latter method is used to parse and insert the data
     * into the given <code>dataStructure</code>.
     */
    protected void register(String type)
    {
	registerChecker(type, "check_"+type );
	registerParser(type, "parse_"+type );
    }

    /**
     * Same as <code>register(String)</code> except one can specify 
     * the method names of the data structure type checker and the parser.
     * This method is mainly useful when the same checker and/or parser is 
     * used to process more than one type.  This can be done by registering
     * the same method for the different types.
     */
    protected void register(String type, String checker, String parser)
    {
	registerChecker(type, checker );
	registerParser(type, parser );
    }
    
    private void registerChecker(String s, String methodName)
    {
	try {
	    checkMap.put(s, 
			 TYPE.getDeclaredMethod(methodName, 
						checkHandlerArgType));
	}
	catch (NoSuchMethodException e) {
	    Util.assert_nonjava(false, "ParseHandlerMap::registerChecker() "+
			"failed to implement method: "+methodName);
	}
    }
    
    private void registerParser(String s, String methodName)
    {
	try {
	    parseMap.put(s, 
			 TYPE.getDeclaredMethod(methodName, 
						parseHandlerArgType));
	}
	catch (NoSuchMethodException e) {
	    Util.assert_nonjava(false, "ParseHandlerMap::registerParser() "+
			"failed to implement method: "+methodName);
	}
    }

    /**
     * A test invoked by Field to see if a particular ParseHandlerMap
     * contain a checker and a parser for a certain field type.
     */
    public boolean understands(String type) 
    {
	return checkMap.containsKey(type) && parseMap.containsKey(type);
    }
    
    /**
     * Gets the registered checker method for the specified <code>type</code>
     * and invokes it on the specified dataStructure.  Returns the
     * result of the type checker method.
     */
    public boolean structCheck(String type, Object dataStructure)
    {
	Object[] tcArgs = { dataStructure };

	Method tc = getTargetDataTypeChecker(type);

	Util.assert_nonjava(tc != null, "ParseHandlerMap::structCheck()");

	boolean result = false;
	try {
	    result = ((Boolean) tc.invoke(this, tcArgs)).booleanValue();
	}
	catch (IllegalAccessException e) {
	    Util.assert_nonjava(false, 
			"ParseHandlerMap::structCheck() illegal access");
	}
	catch (InvocationTargetException e) {
	    e.printStackTrace();
	    Util.assert_nonjava(false, "ParseHandlerMap::structCheck() invoc "+type);
	}
	
	return result;
    }

    /**
     * Gets the registered parser method for the specified <code>type</code>
     * and invokes it on the specified dataStructure and value.  Returns the
     * result of the parser method.
     */
    public void parse(String type, Object dataStructure, String value)
    {
	//System.out.println("getting parse handler");
	Method ph = getParseHandler(type);

	//System.out.println("got parse handler"+ph);

	Util.assert_nonjava(ph != null,
		    "ParseHandlerMap:parse() missing parse handler: "+
		    type);
		
	Object[] phArgs = { dataStructure, value };
	
	try {
	    //	    System.out.println("invoking");
	    ph.invoke(this, phArgs);
	    //	    System.out.println("invoked");
	}
	catch (IllegalAccessException e) {
	    Util.assert_nonjava(false, "ParseHandlerMap::parse() illegal access");
	}
	catch (InvocationTargetException e) {
	    e.printStackTrace();
	    Util.assert_nonjava(false, "ParseHandlerMap::parse() invoc "+type+" "+
			value);
	}
    }
}


